#pragma once

#include <iostream>
#include <fstream>
#include <vector>
#include <map>
#include <queue>
#include <time.h>
#include "ZBDD.h"
#include "SBDD_helper.h"
#include "tdzdd/DdSpec.hpp"
#include "tdzdd/DdStructure.hpp"
#include "tdzdd/spec/SizeConstraint.hpp"
#include "tdzdd/eval/ToZBDD.hpp"
#include "tdzdd/spec/SapporoZdd.hpp"

#include "pathdecompose_frontier/graph.cpp"
using namespace tdzdd;
using namespace sbddh;

long long getCardinarityVecZDD(std::vector<ZBDD> zbdd){
    long long res = 0;
    for(int i = 0; i < zbdd.size(); i++) res += (long long) zbdd[i].Card() * (1LL << i);
    return res;
};

void Init(std::vector<std::vector<int> > &g, int t, std::map<std::tuple<int,int,int>, std::vector<ZBDD> > &allPairMap){
    for(int i = 0; i < g[t].size(); i++){
        std::vector<ZBDD> res;
        res.push_back(getSingleSet(2, t+1, g[t][i] + 1));
        allPairMap[std::make_tuple(g[t][i], t, 1)] = res;
    }
}

void hamilton_dp(Graph G){
    int n = G.numVertices();
    int m = G.numEdges();
    
    std::vector<std::vector<int> > adjacency_list(n);
    std::vector<std::vector<std::pair<int,int> > > g = G.getGraph();
    BDD_Init(1024, 1024 * 1024 * 1024);
    for(int i = 0; i <= n; i++) BDD_NewVar();
    
    for(int v = 0; v < n; v++){
        for(auto adjacency_vertex: g[v]){
            adjacency_list[v].push_back(adjacency_vertex.first);
        }
    }

    int L = G.numLength();

    long long sum = 0;
    
    for(int T = 0; T < n; T++){
        std::map<std::tuple<int,int,int>, std::vector<ZBDD> > allpairMap;
        Init(adjacency_list, T, allpairMap);
        for(int l = 2; l <= L; l++){
            for(int i = 0; i < n; i++){
                for(int j = 0; j < adjacency_list[i].size(); j++){
                    if(allpairMap.count(std::make_tuple(adjacency_list[i][j], T, l - 1))){
                        std::vector<ZBDD> res;
                        if(allpairMap.count(std::make_tuple(i, T, l))) res = allpairMap[std::make_tuple(i, T, l)];
                        
                        auto nextVecZDD = allpairMap[std::make_tuple(adjacency_list[i][j], T, l - 1)];
                        int compsize = std::max(res.size(), nextVecZDD.size());
                        res.resize(compsize);
                        nextVecZDD.resize(compsize);

                        for(int k = 0; k < compsize; k++){
                            auto nextZDD = nextVecZDD[k].OffSet(i+1).Change(i+1);
                            auto intersecZDD = nextZDD & res[k];
                            if(intersecZDD == ZBDD(0)){
                                res[k] += nextZDD;
                            }else{
                                if(k == compsize - 1) res.resize(compsize + 1);
                                res[k] += nextZDD;
                                res[k] -= intersecZDD;
                                
                                int m = k + 1;
                                while(m < res.size() && (res[m] & intersecZDD) != ZBDD(0)){
                                    auto intersec2ZDD = intersecZDD & res[m];
                                    res[m] += intersecZDD;
                                    res[m] -= intersec2ZDD;
                                    intersecZDD = intersec2ZDD;
                                    m++;
                                }

                                if(m == res.size()){
                                    compsize = res.size();
                                    res.resize(compsize + 1); 
                                }

                                res[m] += intersecZDD;
                            }
                        }

                        allpairMap[std::make_tuple(i, T, l)] = res;
                    }
                }
            }
        }

        for(int len = 1; len <= L; len++){
            for(int S = T + 1; S < n; S++){
                if(allpairMap.count(std::make_tuple(S, T, len))){
                    sum += getCardinarityVecZDD(allpairMap[std::make_tuple(S, T, len)]);
                }
            }
        }
    }
    std::cout << sum << std::endl;
}