#pragma once

#include "tdzdd/DdSpec.hpp"
#include "tdzdd/DdSpecOp.hpp"
#include "tdzdd/DdStructure.hpp"
#include "tdzdd/DdEval.hpp"

#include "tdzdd/spec/SizeConstraint.hpp"
#include "tdzdd/util/IntSubset.hpp"

#include "../graph.cpp"
#include "convertToEdgePermutation.cpp"

class PathZDD: public tdzdd::PodArrayDdSpec<PathZDD, short, 2> {
    Graph const G;
    std::vector<int> const perm;
    // std::vector<std::vector<int>> mateInd;
    std::vector<int> permInv;
    std::vector<std::pair<int, int>> edgePermutation;
    std::vector<int> beginInd;
    std::vector<int> lastInd;
    std::vector<std::vector<int>> frontier;
    std::vector<int> stateInd;
    std::vector<std::vector<int>> enteringVertex;
    std::vector<std::vector<int>> leavingVertex;
    int maxMateSize;
    // int stateSize;

public:
    PathZDD(Graph G, std::vector<int> perm): G(G), perm(perm) {
        int n = G.numVertices();
        int m = G.numEdges();
        permInv.resize(n);
        for(int i = 0; i < n; ++i) permInv[perm[i]] = i;

        edgePermutation = convertToEdgePermutation(G, perm);

        beginInd.assign(n, -1);
        lastInd.assign(n, 0);
        for(int i = 0; i < m; ++i) {
            int u = edgePermutation[i].first;
            int v = edgePermutation[i].second;
            if(beginInd[u] == -1) beginInd[u] = i;
            if(beginInd[v] == -1) beginInd[v] = i;
            lastInd[u] = i;
            lastInd[v] = i;
        }

        frontier.resize(m);
        for(int i = 0; i < n; i++) {
            for(int j = beginInd[i]; j <= lastInd[i]; j++) {
                frontier[j].push_back(i);
            }
        }

        maxMateSize = 0;
        for(int i = 0; i < m; i++) {
            if(maxMateSize < frontier[i].size()) maxMateSize = frontier[i].size();
        }

        enteringVertex.resize(m);
        leavingVertex.resize(m);
        for(int i = 0; i < n; ++i) {
            enteringVertex[beginInd[i]].push_back(i);
            leavingVertex[lastInd[i]].push_back(i);
        }

        stateInd.assign(n, -1);
        std::vector<int> stateTmp(n, -1);
        for(int i = 0; i < m; ++i) {
            for(auto t: enteringVertex[i]) {
                for(int j = 0; j < n; ++j) {
                    if(stateTmp[j] == -1) {
                        stateTmp[j] = t;
                        stateInd[t] = j;
                        break;
                    }
                }
            }
            if(i == 0) continue;
            for(auto t: leavingVertex[i-1]){
                for(int j = 0; j < n; ++j) {
                    if(stateTmp[j] == t) {
                        stateTmp[j] = -1;
                        break;
                    }
                }
            }
        }

        setArraySize(maxMateSize + 1);
        // std::cout << maxMateSize + 1 << std::endl;
        // for(int i = 0; i < m; ++i){
        //     std::cerr << i << ": ";
        //     for(int j = 0; j < frontier[i].size(); ++j){
        //         std::cerr << frontier[i][j] << " (" << stateInd[frontier[i][j]] << "), ";
        //     }
        //     std::cerr << std::endl;
        // }
    }

    int getRoot(short* mate) const {
        int n = G.numVertices();
        int m = G.numEdges();
        for(int i = 0; i <= maxMateSize; i++) mate[i] = -1;
        return m;
    }

    int getChild(short* mate, int level, int value) const {
        int s = G.getStart();
        int t = G.getTerminal();
        int n = G.numVertices();
        int m = G.numEdges();
        int u = edgePermutation[m - level].first;
        int v = edgePermutation[m - level].second;
        // int u = G.getEdge(m - level).first;
        // int v = G.getEdge(m - level).second;
        for(auto e: enteringVertex[m - level]) mate[stateInd[e]] = e;
        if(value == 1) {
            // branch cutting
            if(mate[stateInd[u]] == -1 || mate[stateInd[v]] == -1) return 0;
            if(mate[stateInd[u]] ==  v || mate[stateInd[v]] ==  u) return 0;
            // update mate
            if(mate[stateInd[u]] == u && mate[stateInd[v]] == v) {
                mate[stateInd[u]] = v;
                mate[stateInd[v]] = u;
            }
            // else if(mate[stateInd[mate[stateInd[u]]]] == u && mate[stateInd[v]] == v) {
            else if(mate[stateInd[u]] != u && mate[stateInd[v]] == v) {
                int w = mate[stateInd[u]];
                if(lastInd[w] > m - level) mate[stateInd[w]] = v;
                mate[stateInd[v]] = w;
                mate[stateInd[u]] = -1;
            }
            // else if(mate[stateInd[u]] == u && mate[stateInd[mate[stateInd[v]]]] == v) {
            else if(mate[stateInd[u]] == u && mate[stateInd[v]] != v) {
                int w = mate[stateInd[v]];
                if(lastInd[w] > m - level) mate[stateInd[w]] = u;
                mate[stateInd[u]] = w;
                mate[stateInd[v]] = -1;
            }
            // else if(mate[stateInd[mate[stateInd[u]]]] == u && mate[stateInd[mate[stateInd[v]]]] == v) {
            else if(mate[stateInd[u]] != u && mate[stateInd[v]] != v) {
                int w = mate[stateInd[u]];
                int x = mate[stateInd[v]];
                if(lastInd[w] > m - level) mate[stateInd[w]] = x;
                if(lastInd[x] > m - level) mate[stateInd[x]] = w;
                mate[stateInd[u]] = -1;
                mate[stateInd[v]] = -1;
            }
        }
        for(auto e: leavingVertex[m - level]) {
            if(e == s || e == t) {
                if(mate[stateInd[e]] == e || mate[stateInd[e]] == -1) return 0;
            }
            else if(mate[stateInd[e]] != e && mate[stateInd[e]] != -1) return 0;
            mate[stateInd[e]] = -1;
        }
        level--;
        if(level == 0) return -1;
        return level;
    }
};
