#pragma once

#include <tdzdd/DdSpec.hpp>
#include <tdzdd/DdSpecOp.hpp>
#include <tdzdd/DdStructure.hpp>
#include <tdzdd/DdEval.hpp>

#include <tdzdd/spec/SizeConstraint.hpp>
#include <tdzdd/util/IntSubset.hpp>

#include "../graph.cpp"
#include "convertToEdgePermutation.cpp"

class PathZDD1: public tdzdd::PodArrayDdSpec<PathZDD1, char, 2> {
    Graph const G;
    std::vector<int> const perm;
    // std::vector<std::vector<int>> mateInd;
    std::vector<int> permInv;
    std::vector<std::pair<int, int>> edgePermutation;
    std::vector<int> beginInd;
    std::vector<int> lastInd;
    std::vector<std::vector<int>> frontier;
    std::vector<int> stateInd;
    std::vector<std::vector<int>> enteringVertex;
    std::vector<std::vector<int>> leavingVertex;
    int maxMateSize;
    // int stateSize;

public:
    PathZDD1(Graph G, std::vector<int> perm): G(G), perm(perm) {
        int n = G.numVertices();
        int m = G.numEdges();
        permInv.resize(n);
        for(int i = 0; i < n; ++i) permInv[perm[i]] = i;

        edgePermutation = convertToEdgePermutation(G, perm);

        beginInd.assign(n, -1);
        lastInd.assign(n, 0);
        for(int i = 0; i < m; ++i) {
            int u = edgePermutation[i].first;
            int v = edgePermutation[i].second;
            if(beginInd[u] == -1) beginInd[u] = i;
            if(beginInd[v] == -1) beginInd[v] = i;
            lastInd[u] = i;
            lastInd[v] = i;
        }

        frontier.resize(m);
        for(int i = 0; i < n; i++) {
            for(int j = beginInd[i]; j <= lastInd[i]; j++) {
                frontier[j].push_back(i);
            }
        }

        maxMateSize = 0;
        for(int i = 0; i < m; i++) {
            if(maxMateSize < frontier[i].size()) maxMateSize = frontier[i].size();
        }

        enteringVertex.resize(m);
        leavingVertex.resize(m);
        for(int i = 0; i < n; ++i) {
            enteringVertex[beginInd[i]].push_back(i);
            leavingVertex[lastInd[i]].push_back(i);
        }

        stateInd.assign(n, -1);
        std::vector<int> stateTmp(n, -1);
        for(int i = 0; i < m; ++i) {
            for(auto t: enteringVertex[i]) {
                for(int j = 0; j < n; ++j) {
                    if(stateTmp[j] == -1) {
                        stateTmp[j] = t;
                        stateInd[t] = j;
                        break;
                    }
                }
            }
            if(i == 0) continue;
            for(auto t: leavingVertex[i-1]){
                for(int j = 0; j < n; ++j) {
                    if(stateTmp[j] == t) {
                        stateTmp[j] = -1;
                        break;
                    }
                }
            }
        }

        setArraySize(maxMateSize + 1);
        // std::cout << maxMateSize + 1 << std::endl;
        // for(int i = 0; i < m; ++i){
        //     std::cerr << i << ": ";
        //     for(int j = 0; j < frontier[i].size(); ++j){
        //         std::cerr << frontier[i][j] << " (" << stateInd[frontier[i][j]] << "), ";
        //     }
        //     std::cerr << std::endl;
        // }
    }

    int getRoot(char* mate) const {
        int n = G.numVertices();
        int m = G.numEdges();
        for(int i = 0; i <= maxMateSize; i++) mate[i] = -1;
        return m;
    }

    int getChild(char* mate, int level, int value) const {
        int s = G.getStart();
        int t = G.getTerminal();
        int n = G.numVertices();
        int m = G.numEdges();
        int u = edgePermutation[m - level].first;
        int v = edgePermutation[m - level].second;
        for(auto e: enteringVertex[m - level]) mate[stateInd[e]] = 0;
        if(value == 1) {
            if(mate[stateInd[u]] == 2 || mate[stateInd[v]] == 2) return 0;
            mate[stateInd[u]] += 1;
            mate[stateInd[v]] += 1;
        }
        for(auto e: leavingVertex[m - level]) {
            if(e == s || e == t) {
                if(mate[stateInd[e]] != 1) return 0;
            }
            else if(mate[stateInd[e]] == 1) return 0;
            mate[stateInd[e]] = -1;
        }
        level--;
        if(level == 0) return -1;
        return level;
    }
};
